/*
* The following structures are nested within the structures used by the
* system management routines. These structures and sizes were pulled from the AIX
* src tree.
*/
#define LVM_MAXLPS      65535       /* max number of logical partitions allowed */
#define LVM_NAMESIZ     64          /* maximum size for the logical volume name */
#define LVM_NUMCOPIES   3           /* max number of copies allowed of a logical partition */
#define LVM_MAXVGS      255
#define LVM_MAXPVS      32
#define LVM_MAXLVS      256
#define AIX_MIN_BLOCK_SIZE 4096
#define VGSA_BT_PV      127
#define NBPI            32
#define TRUE             1
#define OFFSET_CONSTANT     144
#define SLEEP_TIME            0
#define MAXLVS_OFFSET        16
#define PHYS_VOL_OFFSET      34
#define AIX_PVHPP_LENGTH     PHYS_VOL_OFFSET
#define MAX_SECTORS_NAMELIST 32
#define AIX_DEFAULT_MIRRORING 1
#define AIX_FIRST_MIRROR      2
#define AIX_MAX_MIRRORS       3  // AIX defines ALL copies as mirrors - 3 mirrors MAX - 1 orig and 2 copies

#define EVMS_AIX_FEATURE_ID   3

#define EVMS_AIX_RESYNC_MIRRORS 1

#define PSN_LVM_REC      7
#define PSN_VGSA_REC     128
#define PSN_NAMELIST_REC 2065
#define PSN_VGT_TRAILER  135
#define PSN_LVE_REC        1
#define PSN_PPH_OFFSET    17
#define PSN_PVH_INCREMENT 17
#define AIX_MIN_PVH_SIZE  271  // used to find the PV header info for Pv's other than 0
#define AIX_SECTOR_SIZE  512
#define MAX_PPENT_SECTOR  16
#define NAME_LEN         128    /* don't change!!! */
#define UUID_LEN          32    /* don't change!!! */
#define MAX_SECTORS_LV_ENTRIES 16
#define AIX_MIN_MIRROR_POOL    10
#define AIX_MIRROR_POOL_CHANGE 10

#define LV_SET_ACCESS           _IOW ( 0xfe, 0x28, 1)
#define LV_SET_ALLOCATION       _IOW ( 0xfe, 0x29, 1)
#define LV_SET_STATUS           _IOW ( 0xfe, 0x2a, 1)
#define LV_BMAP 	        _IOWR ( 0xfe, 0x30, 1)

#define LV_ACTIVE            0x01   /* lv_status */
#define LV_SPINDOWN          0x02   /*     "     */
#define LV_ERROR             0x99   /*     "     */

#define VG_ACTIVE            0x01   /* vg_status */

#define AIX_LV_READ          0x00   /* lv_access */
#define AIX_LV_WRITE         0x01   /*     "     */
#define EVMS_LV_NEW          0x10   // volume was created during the current discovery pass
#define EVMS_LV_INCOMPLETE   0x20   // volume has an incomplete LE map
#define EVMS_LV_INVALID      0x40   // volume has a memory-corruption problem

/* vg flags */
#define AIX_VG_DIRTY         0x01   // group has had a new PV added during this discovery
#define AIX_VG_INCOMPLETE    0x20   // volume group is incomplete

#define AIX_LVM_LVUNDEF     0   /* the logical volume is not defined to a */
/* volume group */
#define AIX_LVM_LVDEFINED   1   /* the logical volume is defined to a */
/* volume group */
#define AIX_LVM_LVSTALE     2   /* the logical volume has stale logical */
/* partitions */
#define AIX_LVM_LVMIRBKP    4       /* the logical volume is an online mirror backup */
/* We are skipping '3' since it is used by CMDLVM_LVSTALE */
/* as an addition of LVM_LVDEFINE + LVM_LVSTALE, and is */
/* defined in src/bos/usr/sbin/lvm/include/ls.h */



#define LOG_PREFIX      "--AIXlvm: "

// Entries in the list of physical volumes (PV)
// in a volume group (VG)

struct unique_id {
	u32  word1;
	u32  word2;
	u32  word3;
	u32  word4;
};

struct partition_list_entry {
	struct evms_logical_node     * logical_node;
	u32                pv_number;
	u32                block_size;	 // bytes
	u32                hard_sect_size;   // bytes
	struct partition_list_entry    * next;

};

// Table for mapping logical extents (LE) to physical extents (PE)
struct pe_table_entry {
	struct partition_list_entry  * owning_pv;
	u64               pe_sector_offset;
	char                    pp_state;
};

// Logical volumes (LV) in a volume group (VG)
struct aix_logical_volume {
	u32                lv_number;
	u64                lv_size;	   // Sectors
	u32                lv_access;	   // Flags: LV_READ, LV_WRITE, LN_NEW
	u32                lv_status;	   // Flags: LV_ACTIVE, LV_SPINDOWN
//	u32                lv_minor;	   // Device minor number
	u32                mirror_copies;  // Do we have mirroring and how many  ?
//    u32                mirror_number;  // mirror number - which copy is this ?
//	u32                mirror_iterations;  // Which mirror should we be writing to ?
	u32                stripes;
	u32                stripe_size;	       // Sectors
	u32                stripe_size_shift;  // Number of bits to shift right instead of dividing by stripe_size
	u32                pe_size;	   // Sectors
	u32                pe_size_shift;      // Number of bits to shift right instead of dividing by pe_size
	u32                num_le;	   // Number of entries in the le_to_pe_map
//    u32                new_volume;     // Flag to indicate if this volume needs to be exported
	struct aix_volume_group  * group;      // Pointer back to parent volume group
	unsigned char           name[EVMS_VOLUME_NAME_SIZE+1];	// Dev-tree volume name (eg: /dev/group0/vol0)
	struct pe_table_entry        * le_to_pe_map; // Mapping of logical to physical extents
	struct pe_table_entry        * le_to_pe_map_mir1;    // Mapping of logical to physical extents for mirror 1
	struct pe_table_entry        * le_to_pe_map_mir2;    // Mapping of logical to physical extents for mirror 2
	struct evms_logical_node     * volume_node;  // Pointer to the parent EVMS node representing this volume

};

// Volume groups (VG)
struct aix_volume_group {
	struct unique_id        vg_id;	      // volume group number */
	struct partition_list_entry  * partition_list;	 // List of partitions/segments/PVs that make up this VG
	struct aix_logical_volume      ** volume_list;	 // Array of volumes found in this VG.
	struct aix_volume_group * next;	      // Pointer to the next VG
	struct vg_header       * AIXvgh;      // Pointer to valid data area on disk for the VG
	s32                 vgda_psn;     // Which VGDA we should use
//    u32                numpvs;          // Number of PVs found on this VG.
	u32                numlvs;	      // Number of LVs found on this VG.
	u32                hard_sect_size;    // The largest hard_sect_size and block_size
	u32                block_size;	      // values of all partitions in this group.
	u32                flags;	      //
//    u32                lv_max;          // maximum logical volumes */
	u32                pe_size;	      // physical extent size in sectors */
	u32                partition_count;   // actual partitions found for this group
	u32                CleanVGInfo;	      // Do we have a clean VG Info to work with ?
	u32                vgda_len;	      // length of the volume group descriptor area */
};

struct aix_resync_struct {
	u64                  master_offset;
	u64                  slave1_offset;
	u64                  slave2_offset;
	struct partition_list_entry  * master_part;  //
	struct partition_list_entry  * slave1_part;  //
	struct partition_list_entry  * slave2_part;  //
	struct aix_logical_volume    * resync_vol;
	struct aix_logical_volume    * next_resync_vol;
};

struct aix_mirror_bh {
	atomic_t                     remaining;
	s32                      iteration;	// 'have we finished' count, used from IRQ handlers
	u32                     le;	       // In case we have to flag this pp as stale later.
	s32                      cmd;
	u64                     mir_sector1;
	u64                     mir_sector2;
	struct buffer_head          *master_bh;
	struct buffer_head           bh_req;
	struct aix_mirror_bh        *mirror_bh_list;
	struct evms_logical_node    *node;	    // map to evms node (READ only)
	struct evms_logical_node    *mir_node1;		//
	struct evms_logical_node    *mir_node2;		//
	struct aix_mirror_bh        *next_r1;	    // next for retry or in free list
	char                         sync_flag;	    // Flag for resyncing of mirrored PPs
};

struct aix_volume_resync_ioctl {
	char         object_name[EVMS_VOLUME_NAME_SIZE+1];   // Input  - Name of bbr object from feature header
	s32      force;
};

struct timestruc {
	int tv_sec;
	int tv_nsec;

};

struct aix_ipl_rec_area {
	u32          IPL_record_id;    /* This physical volume contains a   */
	/* valid IPL record if and only if   */
	/* this field contains IPLRECID      */

#define IPLRECID 0xc9c2d4c1             /* Value is EBCIDIC 'IBMA'           */

	char              reserved1[20];
	u32          formatted_cap;    /* Formatted capacity. The number of */
	/* sectors available after formatting*/
	/* The presence or absence of bad    */
	/* blocks does not alter this value. */

	char              last_head;	    /* THIS IS DISKETTE INFORMATION      */
					    /* The number of heads minus 1. Heads*/
					    /* are number from 0 to last_head.   */

	char              last_sector;	    /* THIS IS DISKETTE INFORMATION      */
					    /* The number of sectors per track.  */
					    /* Sectors are numbered from 1 to    */
					    /* last_sector.                      */

	char              reserved2[6];

	u32          boot_code_length; /* Boot code length in sectors. A 0  */
	/* value implies no boot code present*/

	u32          boot_code_offset; /* Boot code offset. Must be 0 if no */
	/* boot code present, else contains  */
	/* byte offset from start of boot    */
	/* code to first instruction.        */

	u32          boot_lv_start;    /* Contains the PSN of the start of  */
	/* the BLV.                          */

	u32          boot_prg_start;   /* Boot code start. Must be 0 if no  */
	/* boot code present, else contains  */
	/* the PSN of the start of boot code.*/

	u32          boot_lv_length;   /* BLV length in sectors.            */

	u32          boot_load_add;    /* 512 byte boundary load address for*/
	/* boot code.                        */

	char              boot_frag;	    /* Boot code fragmentation flag. Must*/
					    /* be 0 if no fragmentation allowed, */
					    /* else must be 0x01.                */

	char          boot_emulation;	/* ROS network emulation flag */
	/* 0x0 => not an emul support image   */
	/* 0x1 => ROS network emulation code  */
	/* 0x2 => AIX code supporting ROS emul*/

	char              reserved3[2];

	u16          basecn_length;    /* Number of sectors for base        */
	/* customization. Normal mode.       */

	u16          basecs_length;    /* Number of sectors for base        */
	/* customization. Service mode.      */

	u32          basecn_start;     /* Starting PSN value for base       */
	/* customization. Normal mode.       */

	u32          basecs_start;     /* Starting PSN value for base       */
	/* customization. Service mode.      */

	char              reserved4[24];

	u32          ser_code_length;  /* Service code length in sectors.   */
	/* A 0 value implies no service code */
	/* present.                          */

	u32          ser_code_offset;  /* Service code offset. Must be 0 if */
	/* no service code is present, else  */
	/* contains byte offset from start of*/
	/* service code to first instruction.*/

	u32          ser_lv_start;     /* Contains the PSN of the start of  */
	/* the SLV.                          */

	u32          ser_prg_start;    /* Service code start. Must be 0 if  */
	/* service code is not present, else */
	/* contains the PSN of the start of  */
	/* service code.                     */

	u32          ser_lv_length;    /* SLV length in sectors.            */

	u32          ser_load_add;     /* 512 byte boundary load address for*/
	/* service code.                     */

	char              ser_frag;	    /* Service code fragmentation flag.  */
					    /* Must be 0 if no fragmentation     */
					    /* allowed, else must be 0x01.       */

	char          ser_emulation;	/* ROS network emulation flag */
	/* 0x0 => not an emul support image   */
	/* 0x1 => ROS network emulation code  */
	/* 0x2 => AIX code supporting ROS emul*/

	char              reserved5[2];

	struct unique_id  pv_id;	    /* The unique identifier for this    */
					    /* physical volume.                  */
	char              dummy[512 - 128 - sizeof(struct unique_id)];
};


struct AIXlvm_rec
/* structure which describes the physical volume LVM record */ {
	u32           lvm_id;	       /* LVM id field which identifies whether the PV is a member of a volume group */

#define AIX_LVM_LVMID     0x5F4C564D            /* LVM id field of ASCII "_LVM" */

	struct unique_id          vg_id;	   /* the id of the volume group to which this physical volume belongs */
	u32           lvmarea_len;     /* the length of the LVM reserved area */
	u32           vgda_len;	       /* length of the volume group descriptor area */
	s32            vgda_psn [2];    /* the physical sector numbers of the beginning of the volume group descriptor area copies on this disk */
	s32            reloc_psn;	    /* the physical sector number of the beginning of a pool of blocks  */
					    /* (located at the end of the PV) which are reserved for the relocation of bad blocks */
	u32           reloc_len;       /* the length in number of sectors of the pool of bad block relocation blocks */
	s16            pv_num;		/* the physical volume number within the volume group of this physical volume */
	s16            pp_size;		/* the size in bytes for the partition, expressed as a power of 2 (i.e., the partition size is 2 to the power pp_size) */
	u32           vgsa_len;	       /* length of the volume group status area */
	s32            vgsa_psn [2];    /* the physical sector numbers of the beginning of the volume group status area copies on this disk */
	s16            version;		/* the version number of this volume group descriptor and status area */

#define  LVM_VERSION_1      1              /* first version - AIX 3.0 */
#define  LVM_STRIPE_ENHANCE 2              /* version with striped lv's - AIX 4.1 */
#define  LVM_1024_PPSIZE    3              /* ppsizes of 512 and 1024 */
#define  LVM_GT_1016        4              /* version with support for > 1016 pps/pv */
#define  LVM_MAX_VERSION    LVM_GT_1016    /* max version # */

	char res1 [450];		    /* reserved area */

};



/*  II.Volume Group Descriptor Area  */

struct vgsa_area {
	struct timestruc           b_tmstamp;	 /* Beginning timestamp */
	u32          pv_missing [(LVM_MAXPVS + (NBPI -1)) / NBPI];  /* Bit per PV */
	unsigned char         stalepp    [LVM_MAXPVS] [VGSA_BT_PV];
	s16               factor;
	char                  resv[10];	    /* Padding */
	struct timestruc           e_tmstamp;	 /* Ending timestamp */

} ;

struct vg_header {
	struct timestruc           vg_timestamp; /* time of last update */
	struct unique_id             vg_id;	   /* unique id for volume group */
	s16               numlvs;	/* number of lvs in vg */
	s16               maxlvs;	/* max number of lvs allowed in vg */
	s16               pp_size;	/* size of pps in the vg */
	s16               numpvs;	/* number of pvs in the vg */
	s16               total_vgdas;	/* number of copies of vg */
	/* descriptor area on disk */
	s16               vgda_size;	/* size of volume group descriptor */
	s16               bigvg;
	s16               quorum;
	s16               auto_varyon;
	s32               checksum;
	s32               bigda_size;
};

struct lv_entries {
	s16               lvname;	  /* name of LV */
	s16               res1;		  /* reserved area */
	s32               maxsize;	  /* maximum number of partitions allowed */
	char                  lv_state;	      /* state of logical volume */
	char                  mirror;	      /* none,single, or double */
	s16               mirror_policy;  /* type of writing used to write */
	s32               num_lps;	  /* number of logical partitions on the lv */
	/* base 1 */
	char                  permissions;    /* read write or read only */
	char                  bb_relocation;  /* specifies if bad block */
					      /* relocation is desired */
	char                  write_verify;   /* verify all writes to the LV */
	char                  mirwrt_consist; /* mirror write consistency flag */
	u16              stripe_exp;	 /* stripe size in exponent value */
	u16              striping_width; /* stripe width */
	u16              lv_avoid;
	u16              child_minor_num;
	char                  res4[4];	      /* reserved area on disk */
};


struct pv_header {
	struct unique_id             pv_id;	 /* unique identifier of PV */
	u16              pp_count;   /* number of physical partitions */
	/* on PV */
	char                  pv_state;	  /* state of physical volume */
	char                  res1;	  /* reserved area on disk */
	s32               psn_part1;  /* physical sector number of 1st pp */
	s16               pvnum_vgdas;/* number of vg descriptor areas */
	/* on the physical volume */
	s16               pv_num;     /* PV number */
	u32              res2;	     /* reserved area on disk */

};

struct pp_entries {
	s16              lv_index;     /* index to lv pp is on */
	s16              res_1;	       /* reserved area on disk */
	u32             lp_num;	      /* log. part. number */
	char                 copy;	   /* the copy of the logical partition */
					   /* that this pp is allocated for */
	char                 pp_state;	   /* current state of pp */
	char                 fst_alt_vol;  /* pv where partition allocation for*/
					   /* first mirror begins */
	char                 snd_alt_vol;  /* pv where partition allocation for*/
					   /* second mirror begins */
	s16              fst_alt_part; /* partition to begin first mirror */
	s16              snd_alt_part; /*partition to begin second mirror */
	u64             res_3;	      /* reserved area  on disk */
	u64             res_4;	      /* reserved area on disk */
};

struct namelist {
	char       name[LVM_MAXLVS][LVM_NAMESIZ];
};

struct vg_trailer {
	struct timestruc     timestamp;	/*  time of last update */
	s16         concurrency;
	/* MS Nibble = concurrent capable           */
	/* LS Nibble = concurrent auto-varyon           */
	s16         res_2;
	s32         res_3;  /* reserved area on disk */
	u64        res_4;  /* reserved area on disk */
	u64        res_5;  /* reserved area on disk */
};

