/* -*- linux-c -*- */
/*
 *   Copyright (c) International Business Machines  Corp., 2000
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or 
 *   (at your option) any later version.
 * 
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software 
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
/*
 * linux/include/linux/evms_lvm.h
 *
 * EVMS LVM VGE kernel header file
 */

#ifndef __EVMS_LVM_H__
#define __EVMS_LVM_H__

#define EVMS_LVM_VERSION_MAJOR	1
#define EVMS_LVM_VERSION_MINOR	1
#define EVMS_LVM_VERSION_PATCH	2

/* The following definitions and data structures are copied from lvm.h and
 * liblvm.h from the LVM 0.9.1beta8 distribution. Since the metadata format
 * changed in beta8, lvm.h changed significantly enough that this module would
 * no longer compile. Instead of requiring evms users to install the latest lvm 
 * release, the required definitions and data structures will now be included
 * in this header file.
 */

#define MAX_VG  		99
#define MAX_LV			256
#define	MAX_PV			256
#define	NAME_LEN		128
#define	UUID_LEN		32
#define LVM_VGDA_ALIGN		4096UL
#define	LVM_PV_DISK_BASE	0L
#define	LVM_PV_DISK_SIZE	1024L
#define	LVM_VG_DISK_BASE	round_up(LVM_PV_DISK_BASE + LVM_PV_DISK_SIZE, \
					 LVM_VGDA_ALIGN)
#define	LVM_VG_DISK_SIZE  	(8*512L)

/*
 * Status flags
 */
/* lv->lv_status */
#define	LV_ACTIVE		0x01
/* lv->lv_access */
#define	LV_READ			0x01
#define	LV_WRITE		0x02
#define	LV_SNAPSHOT		0x04
#define	LV_SNAPSHOT_ORG		0x08

/**
 * struct lv_COW_table_disk_v1
 * @pv_org_number:
 * @pv_org_rsector:
 * @pv_snap_number:
 * @pv_snap_rsector:
 *
 * Copy-On-Write tables in disk format (version 1).
 **/
struct lv_COW_table_disk {
	u64 pv_org_number;
	u64 pv_org_rsector;
	u64 pv_snap_number;
	u64 pv_snap_rsector;
};

/**
 * struct pe_disk
 * @lv_num:
 * @le_num:
 *
 * Disk stored PE map entry definition.
 **/
struct pe_disk {
	u16 lv_num;
	u16 le_num;
};

/**
 * struct lvm_disk_data
 * @base:
 * @size:
 *
 * Disk stored PV, VG, LV and PE size and offset information.
 */
struct lvm_disk_data {
	u32 base;
	u32 size;
};

/**
 * struct pv_disk
 * @id:
 * @version:
 * @pv_on_disk:
 * @vg_on_disk:
 * @pv_uuidlist_on_disk:
 * @lv_on_disk:
 * @pe_on_disk:
 * @pv_uuid:
 * @vg_name:
 * @system_id: used by vgexport/vgimport
 * @pv_major:
 * @pv_number:
 * @pv_status:
 * @pv_allocatable:
 * @pv_size:
 * @lv_cur
 * @pe_size:
 * @pe_total:
 * @pe_allocated:
 * @pe_start: in sectors (new in version 2)
 *
 * Physical volume on disk metadata definition (version 2).
 */
struct pv_disk {
	u8			id[2];
	u16			version;
	struct lvm_disk_data	pv_on_disk;
	struct lvm_disk_data	vg_on_disk;
	struct lvm_disk_data	pv_uuidlist_on_disk;
	struct lvm_disk_data	lv_on_disk;
	struct lvm_disk_data	pe_on_disk;
	u8			pv_uuid[NAME_LEN];
	u8			vg_name[NAME_LEN];
	u8			system_id[NAME_LEN];
	u32			pv_major;
	u32			pv_number;
	u32			pv_status;
	u32			pv_allocatable;
	u32			pv_size;
	u32			lv_cur;
	u32			pe_size;
	u32			pe_total;
	u32			pe_allocated;
	u32			pe_start;
};

/**
 * struct lv_disk
 * @lv_name:
 * @vg_name:
 * @lv_access:
 * @lv_status:
 * @lv_open:
 * @lv_dev:
 * @lv_number:
 * @lv_mirror_copies:
 * @lv_recovery:
 * @lv_schedule:
 * @lv_size:
 * @lv_snapshot_minor: minor number of original
 * @lv_chunk_size: chuck size for snapshots
 * @lv_dummy:
 * @lv_allocated_le:
 * @lv_stripes:
 * @lv_stripesize:
 * @lv_badblock:
 * @lv_allocated:
 * @lv_io_timeout:
 * @lv_read_ahead:
 *
 * Logical volume metadata definition (version 3).
 */
struct lv_disk {
	u8	lv_name[NAME_LEN];
	u8	vg_name[NAME_LEN];
	u32	lv_access;
	u32	lv_status;
	u32	lv_open;
	u32	lv_dev;
	u32	lv_number;
	u32	lv_mirror_copies;
	u32	lv_recovery;
	u32	lv_schedule;
	u32	lv_size;
	u32	lv_snapshot_minor;
	u16	lv_chunk_size;
	u16	dummy;
	u32	lv_allocated_le;
	u32	lv_stripes;
	u32	lv_stripesize;
	u32	lv_badblock;
	u32	lv_allocation;
	u32	lv_io_timeout;
	u32	lv_read_ahead;
};

/**
 * struct vg_disk
 * @vg_uuid:		Volume group UUID
 * @vg_name_dummy:	Remainder of version 1 VG name
 * @vg_number:		Volume group number
 * @vg_access:		Read/Write
 * @vg_status:		Active or not
 * @lv_max:		Maximum logical volumes
 * @lv_cur:		Current logical volumes
 * @lv_open:		Open logical volumes
 * @pv_max:		Maximum physical volumes
 * @pv_cur:		Current physical volumes
 * @pv_act:		Active physical volumes
 * @dummy:
 * @vgda:		Volume group descriptor arrays
 * @pe_size:		Physical extent size in sectors
 * @pe_total:		Total of physical extents
 * @pe_allocated:	Allocated physical extents
 * @pvg_total:		Physical volume groups
 *
 * Volume group metadata definition (version 2).
 */
struct vg_disk {
	u8	vg_uuid[UUID_LEN];
	u8	vg_name_dummy[NAME_LEN - UUID_LEN];
	u32	vg_number;
	u32	vg_access;
	u32	vg_status;
	u32	lv_max;
	u32	lv_cur;
	u32	lv_open;
	u32	pv_max;
	u32	pv_cur;
	u32	pv_act;
	u32	dummy;
	u32	vgda;
	u32	pe_size;
	u32	pe_total;
	u32	pe_allocated;
	u32	pvg_total;
};

/* Useful inlines */
static inline ulong round_up(ulong n, ulong size)
{
	size--;
	return (n + size) & ~size;
}

static inline ulong div_up(ulong n, ulong size)
{
	return round_up(n, size) / size;
}

/* End of lvm.h imported data structures. */

#define DEV_DIRECTORY		"/dev/"
#define LVM_DEV_DIRECTORY	"lvm/"
#define LVM_PROC_NAME		"lvm"
#define LVM_PROC_VG_NAME	"VGs"
#define LVM_PROC_LV_NAME	"LVs"
#define LVM_PROC_PV_NAME	"PVs"
#define LVM_PROC_GLOBAL_NAME	"global"
#define IO_BUFFER_SECTORS	8

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,4,9)
#define max(a,b) (((a)>(b))?(a):(b))
#endif

/* Structure for doing PV remove ioctls. */

#define EVMS_LVM_PV_REMOVE_IOCTL	0x01
#define EVMS_LVM_SNAPSHOT_STAT_IOCTL	0x02

/**
 * struct lvm_pv_remove_ioctl
 * @vg_uuid:	Volume group UUID
 * @pv_number:	Physical volume number
 * @next:	Link to next packet (engine-use only)
 *
 * PV remove ioctl packet definition.
 */
struct lvm_pv_remove_ioctl {
	u8				vg_uuid[UUID_LEN];
	u32				pv_number;
	struct lvm_pv_remove_ioctl	* next;
};

/**
 * struct lvm_snapshot_stat_ioctl
 * @vg_uuid:		Volume group UUID
 * @lv_number:		Logical volume number
 * @next_free_chuck:
 * @lv_status:
 *
 * Snapshot statistics ioctl packet definition.
 **/
struct lvm_snapshot_stat_ioctl {
	u8	vg_uuid[UUID_LEN];
	u32	lv_number;
	u64	next_free_chunk;
	u32	lv_status;
};

/**
 * struct lvm_physical_volume
 * @logical_node:	Storage object
 * @pv:			Copy of on-disk PV struct
 * @pe_map:
 * @pv_number:
 * @next:		Pointer to next entry
 *
 * Entries in the list of physical volumes (PV) in a volume group (VG).
 */
struct lvm_physical_volume {
	struct evms_logical_node	* logical_node;
	struct pv_disk			* pv;
	struct pe_disk			* pe_map;
	u32				pv_number;
	struct lvm_physical_volume	* next;
};

/**
 * struct le_table_entry
 * @owning_pv:
 * @pe_sector_offset:
 *
 * Table entry definition for mapping logical
 * extents (LE) to physical extents (PE).
 */
struct le_table_entry {
	struct lvm_physical_volume	* owning_pv;
	u64				pe_sector_offset;
};

/**
 * struct snapshot_map_entry
 * @org_sector:
 * @snap_sector:
 * @snap_pv:
 * @next:
 * @prev:
 *
 * Snapshot remapping entry structure definition.
 */
struct snapshot_map_entry {
	u64				org_sector;
	u64				snap_sector;
	struct lvm_physical_volume	* snap_pv;
	struct snapshot_map_entry	* next;
	struct snapshot_map_entry	* prev;
};

#define MAX_HASH_CHAIN_ENTRIES	10
#define CHUNK_DATA_BUFFER_SIZE	128

/**
 * struct lvm_logical_volume
 * @lv_number:
 * @lv_size:		In sectors
 * @lv_access:		Flags: LV_READ, LV_WRITE, LV_SNAPSHOT,
 *				LV_SNAPSHOT_ROG, EVMS_LV*
 * @lv_status:		Flags: LV_ACTIVE, LV_SPINDOWN
 * @lv_minor:		Device minor number
 * @stripes:
 * @stripe_size:	In sectors
 * @stripe_size_shift:	# of bits to shift right instead of dividing by stripe_size
 * @pe_size:		In sectors
 * @pe_size_shift:	Number of bits to shift right instead of dividing by pe_size
 * @num_le:		Number of entries in the le-to-pe map
 * @group:		Pointer back to parent volume group
 * @name:		Dev-tree volume name (eg. /dev/group0/vol0)
 * @le_map:		Mapping of logical to physical extents
 * @volume_node:	Pointer to parent EVMS object representing this volume
 * @chunk_size:		In sectors
 * @num_chunks:		lv_size / chunk_size
 * @snap_org_minor:	Minor number of snapshot original
 * @next_cow_entry:	Index into current COW table
 * @current_cow_sector:	Logical sector of current COW table
 * @next_free_chunk:	Starting logical sector of next free chunk
 * @hash_table_size:	Number of pointers in each hash table
 * @cow_table:		Pointer to one sector's worth of COW tables.
 * @chunk_data_buffer:	Buffer reading data when doing copy-on-write
 * @snap_semaphore:	For locking during snapshot IO operations
 * @snapshot_map:	Pointer to remapping hash tables
 * @snapshot_next:	Linked list of volumes being snapshotted
 * @snapshot_org:	Pointer to volume being snapshotted
 *
 * In-memory representation of an LVM LV.
 */
struct lvm_logical_volume {
	u32			lv_number;
	u64			lv_size;
	u32			lv_access;
	u32			lv_status;
	u32			lv_minor;
	u32			stripes;
	u32			stripe_size;
	u32			stripe_size_shift;
	u32			pe_size;
	u32			pe_size_shift;
	u32			num_le;
	struct lvm_volume_group	* group;
	u8			name[NAME_LEN];
	struct le_table_entry	* le_map;
	struct evms_logical_node * volume_node;
	u32			chunk_size;
	u32			num_chunks;
	u32			snap_org_minor;
	u32			next_cow_entry;
	u64			current_cow_sector;
	u64			next_free_chunk;
	u32			hash_table_size;
	struct lv_COW_table_disk * cow_table;
	u8			* chunk_data_buffer;
	struct semaphore	snap_semaphore;
	struct snapshot_map_entry *** snapshot_map;
	struct lvm_logical_volume * snapshot_next;
	struct lvm_logical_volume * snapshot_org;
};

/* lv_access:
 * EVMS_LV_NEW:		Volume was created during the current discovery pass.
 * EVMS_LV_INCOMPLETE:	Volume has an incomplete LE map.
 * EVMS_LV_INVALID:	Volume has a memory-corruption problem.
 * EVMS_LV_QUIESCED:	Volume is in quiesced state.
 * EVMS_LV_EXPORTED:	Volume has been exported during this EVMS discovery pass.
 */
#define EVMS_LV_NEW		0x10
#define EVMS_LV_INCOMPLETE	0x20
#define EVMS_LV_INVALID		0x40
#define EVMS_LV_QUIESCED	0x80
#define EVMS_LV_EXPORTED	0x100

/**
 * struct lvm_volume_group
 * @vg:			Copy of on-disk VG metadata
 * @pv_list:		List of PVs that make up this group
 * @volume_list:	Array of volumes
 * @lv_array:		Array of LV metadata
 * @lv_array_disk:
 * @uuid_list:		List of PV UUIDs
 * @vg_uuid:		UUID from the VG metadata
 * @vg_name:		Name from the PV metadata
 * @pv_count:		# of PVs found in this group
 * @volume_count:	# of LVs found in this group
 * @hard_sect_size:	Largest hardsector size of all PVs in this group
 * @block_size:		Largest block size of all PVs in this group
 * @flags:		EVMS_VG*
 * @next_group:		Linked list
 *
 * In-memory representation of an LVM VG.
 */
struct lvm_volume_group {
	struct vg_disk			* vg;
	struct lvm_physical_volume	* pv_list;
	struct lvm_logical_volume	* volume_list[MAX_LV + 1];
	struct lv_disk			* lv_array;
	struct lv_disk			* lv_array_disk;
	u8				* uuid_list;
	u8				vg_uuid[UUID_LEN];
	u8				vg_name[NAME_LEN];
	u32				pv_count;
	u32				volume_count;
	s32				hard_sect_size;
	s32				block_size;
	u32				flags;
	struct lvm_volume_group		* next_group;
};

/* flags
 * EVMS_VG_DIRTY:		Group is new or has had a PV added
 *				during this discovery.
 * EVMS_VG_PARTIAL_PVS:		Group contains at least one partial PV.
 * EVMS_VG_REMOVABLE_PVS:	Group contains at least one removable PV.
 */
#define EVMS_VG_DIRTY			(1 << 0)
#define EVMS_VG_PARTIAL_PVS		(1 << 1)
#define EVMS_VG_REMOVABLE_PVS		(1 << 2)

#endif
