/*
 *   Copyright (c) International Business Machines  Corp., 2000
 *
 *   This program is free software;  you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY;  without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program;  if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * linux/include/linux/evms/evms_md.h
 *
 * EVMS Linux MD Region Manager Public Header File
 *
 * 'evms_md.h' is an EVMS version of linux/include/linux/raid/md.h modified
 * by Cuong (Mike) Tran <miketran@us.ibm.com>, January 2002.
 *
 */

#ifndef __EVMS_MD_INCLUDED
#define __EVMS_MD_INCLUDED

#include <linux/mm.h>
#include <linux/fs.h>
#include <linux/blkdev.h>
#include <asm/semaphore.h>
#include <linux/ioctl.h>
#include <linux/types.h>
#include <asm/bitops.h>
#include <linux/module.h>
#include <linux/hdreg.h>
#include <linux/proc_fs.h>
#include <linux/smp_lock.h>
#include <linux/delay.h>
#include <net/checksum.h>
#include <linux/random.h>
#include <linux/locks.h>
#include <linux/kernel_stat.h>
#include <asm/io.h>
#include <linux/completion.h>

#include <linux/evms/evms.h>

#include <linux/raid/md_compatible.h>
/*
 * 'md_p.h' holds the 'physical' layout of RAID devices
 * 'md_u.h' holds the user <=> kernel API
 *
 * 'md_k.h' holds kernel internal definitions
 */

#include <linux/evms/evms_md_p.h>
#include <linux/evms/evms_md_u.h>
#include <linux/evms/evms_md_k.h>

/*
 * Different major versions are not compatible.
 * Different minor versions are only downward compatible.
 * Different patchlevel versions are downward and upward compatible.
 */
#define EVMS_MD_MAJOR_VERSION                1
#define EVMS_MD_MINOR_VERSION                1
#define EVMS_MD_PATCHLEVEL_VERSION           2

#define MD_MAJOR_VERSION                	0
#define MD_MINOR_VERSION                	90
#define MD_PATCHLEVEL_VERSION           	0

#define EVMS_MD_COMMON_SERVICES_MAJOR		0
#define EVMS_MD_COMMON_SERVICES_MINOR		5
#define EVMS_MD_COMMON_SERVICES_PATCHLEVEL	0


extern int evms_md_size[MAX_MD_DEVS];

extern void evms_md_add_mddev_mapping (mddev_t *mddev, kdev_t dev, void *data);
extern void evms_md_del_mddev_mapping (mddev_t *mddev, kdev_t dev);
extern char * evms_md_partition_name (struct evms_logical_node *node);
extern int evms_register_md_personality (int p_num, mdk_personality_t *p);
extern int evms_unregister_md_personality (int p_num);

extern int evms_md_update_sb (mddev_t *mddev);
extern int evms_md_check_ordering (mddev_t *mddev);
extern void evms_md_print_devices (void);

extern int evms_md_sync_io(
	struct evms_logical_node *node,	/* evms node for the MD array */
        int rw,				/* READ / WRITE */
        u64 sector,			/* starting sector */
        u64 total_nr_sects,		/* total number of sectors */
        void *data );			/* pointer to buffer */

extern int evms_md_partial_sync_io(
	struct evms_logical_node *node, /* evms node for the MD array */
	int rw,			/* READ / WRITE */
        u64 sector, 		/* starting sector */
	u32 *nsects,		/* on input: the total number of sectors for the request */
				/* on output, number of sectors completed */
        void *data);		/* pointer to buffer */


extern int evms_md_do_sync(mddev_t *mddev, mdp_disk_t *spare);
extern void evms_md_done_sync(mddev_t *mddev, int blocks, int ok);
extern void evms_md_sync_acct(kdev_t dev, unsigned long nr_sectors);
extern void evms_md_recover_arrays (void);
extern int evms_md_error (mddev_t *mddev, struct evms_logical_node *node);
extern int evms_md_error_dev(mddev_t *mddev, kdev_t dev);

#define MD_BUG(x...) { printk("md: bug in file %s, line %d\n", __FILE__, __LINE__); evms_md_print_devices(); }


#endif 

