 /*
 * INET		An implementation of the TCP/IP protocol suite for the LINUX
 *		operating system.  INET is implemented using the  BSD Socket
 *		interface as the means of communication with the user level.
 *
 *		Global definitions for the Token-Ring IEEE 802.5 interface.
 *
 * Version:	@(#)if_tr.h	0.0	07/11/94
 *
 * Author:	Fred N. van Kempen, <waltje@uWalt.NL.Mugnet.ORG>
 *		Donald Becker, <becker@super.org>
 *              Peter De Schrijver, <stud11@cc4.kuleuven.ac.be>
 *              D.J. Barrow barrow_dj@yahoo.com
 *
 *		This program is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 */
#ifndef _LINUX_IF_TR_H
#define _LINUX_IF_TR_H

/* IEEE 802.5 Token-Ring magic constants.  The frame sizes omit the preamble
   and FCS/CRC (frame check sequence). */
#define TR_ALEN	6		/* Octets in one ethernet addr	 */
#define TR_HLEN   (sizeof(struct trh_hdr)+sizeof(hdr_8022))
#define AC			0x10
#define LLC_FRAME               0x40


/* LLC and SNAP constants */
#define TR_IP_SAP    0xAA
#define TR_IPX_SAP   0xE0
#define EXTENDED_SAP 0xAA
#define UI_CMD       0x03


/* This is an Token-Ring frame header. */
/* access control field */
/* frame control field */
/* destination address */
/* source address */
#define TR_HW_HDR                   \
__u8  ac;		            \
__u8  fc;		            \
__u8  daddr[TR_ALEN];	            \
__u8  saddr[TR_ALEN];

typedef struct 
{
	TR_HW_HDR
} tr_hw_hdr  __attribute__((packed));

typedef struct
{
	__u8  saddr[TR_ALEN];
	unsigned short type;
} tr_fake_header;
#define TR_NUM_RSEGS 8

/* route control field */
/* routing registers */
#define TR_RIF_INFO                                 \
__u16 rcf __attribute__((packed));                  \
__u16 rseg[TR_NUM_RSEGS] __attribute__((packed));   		

typedef struct
{
        TR_RIF_INFO
} tr_rif_info __attribute__((packed));
#define TR_MAXRIFLEN sizeof(tr_rif_info)
#define TR_MAX_NUM_HOPS (TR_NUM_RSEGS-1)



/* Token-Ring statistics collection data. */
struct tr_statistics 
{
	unsigned long rx_packets;       /* total packets received	*/
	unsigned long tx_packets;	/* total packets transmitted	*/
	unsigned long rx_bytes;		/* total bytes received   	*/
	unsigned long tx_bytes;		/* total bytes transmitted	*/
	unsigned long rx_errors;	/* bad packets received		*/
	unsigned long tx_errors;	/* packet transmit problems	*/
	unsigned long rx_dropped;	/* no space in linux buffers	*/
	unsigned long tx_dropped;	/* no space available in linux	*/
	unsigned long multicast;	/* multicast packets received	*/
	unsigned long transmit_collision;

	/* detailed Token-Ring errors. See IBM Token-Ring Network
	   Architecture for more info */

	unsigned long line_errors;
	unsigned long internal_errors;
	unsigned long burst_errors;
	unsigned long A_C_errors;
	unsigned long abort_delimiters;
	unsigned long lost_frames;
	unsigned long recv_congest_count;
	unsigned long frame_copied_errors;
	unsigned long frequency_errors;
	unsigned long token_errors;
	unsigned long dummy1;
};


#define LLC_8022   \
__u8  dsap;        \
__u8  ssap;        \
__u8  llc;       

#define SNAP_8022   \
__u8  protid[3];    \
__u16 ethertype;  

typedef struct
{
     LLC_8022
}  llc_8022  __attribute__((packed));
typedef struct
{
     SNAP_8022
}  snap_8022  __attribute__((packed));

typedef struct
{
        LLC_8022
        SNAP_8022
}  hdr_8022  __attribute__((packed));

/* for backward compatibility mainly */
struct  trh_hdr 
{
        TR_HW_HDR
        TR_RIF_INFO
} __attribute__((packed));

struct trllc
{
        LLC_8022
        SNAP_8022
} __attribute__((packed));

#define TR_RCF_FRAME2K 0x20
/* #define TR_MAXRIFLEN 18 */

/*
  RFC 1749 stuff
 */
/* 0=individual addr 1=group/multicast addr */
#define TR_IG_BIT   0x80
/* FAI bit 0=functional address,1= indicates locally administered group address
*/
#define TR_FAI_BIT  0x80
/* 0=universally administered 1=locally administered */
#define TR_UL_BIT  0x40
/* This isn't needed to set the broadcast bit in rcfs as the TR_IG_BIT is already set */
#define IP_MCAST_FUNC_ADDR {0xC0,0x00,0x00,0x04,0x00,0x00}

/* source routing stuff */
#define TR_RII 0x80
#define TR_RCF_DIR_BIT 0x80
#define TR_RCF_LF_MASK  0x70
#define TR_RCF_LEN_MASK 0x1f00
#define TR_NORMAL                  0x0         /* normal transmit */
#define TR_NO_RCF		   0x1         /* i.e. don't put an rcf in the header */
#define TR_RCF_ALL_RINGS_BROADCAST 0x8000      /* hit every ring possible ring broadcast */
#define TR_RCF_LIMITED_BROADCAST   0xC000      /* hit every ring at most once once broadcast */
#define TR_RCF_BROADCAST_MASK      0xE000
#define TR_RCF_CACHE_MASK     ~(TR_RCF_BROADCAST_MASK)

/* Should be in types.h but ... */
#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif
#define LOCAL_RING -1

/*

  The format of each Routing Descriptor (RD) field is: ( RFC 1749 )

                     octet 1         octet 2
                 +---------------+---------------+
                 |r r r r r r r r r r r r i i i i|
                 +---------------+---------------+
                  <---- ring number ----> <----->
                                             ^
                                             |
                             bridge number --+ // should be bridge type

                      0                   1
                      0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5
                     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
                     |  B  |   LTH   |D|  LF |   r   |
                     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

*/

typedef struct tr_rif_cache_entry tr_rif_cache_entry; 
/* This list structure is fairly complex ( I got a headache writing it )
 * ( & don't get put off by all the debug code in tr_add_rif_info. )
 * It may not add much performance on networks  with less than 100 nodes
 *  but it scales very well.
 * The 2 doubly linked lists nearly completely eliminates any list walking.
 * The idea behind the lists is as follows
 * The most recently used list is needed to help
 * maintain the least recently used list correctly.
 * The least recently used list is used so that tr_add_rif_info
 * dosen't have to walk the entire list to find out
 * which member to replace in the cache when a line is full.
 * The updated list is used because
 * most of the packets received in tr_add_rif_info will be at the
 * top of this list if they received packets from this mac address
 * in the recent past thus avoiding list walking.
 * The hint field added to struct sock means that
 * the no list walking is done on transmitted packets except to 
 * set up the initial hint DJB.
 * obviously the touched list dosent perform quite as well as
 * the lru list as the packets coming in are much more random
 * The hint_idx field is required to tell if a hint is valid
 * this is done by updating an index each time we free a cache entry.
 */
typedef struct
{
	tr_rif_cache_entry *head;
	tr_rif_cache_entry *tail;
} tr_rif_dlist;

typedef struct
{
	tr_rif_cache_entry *next;
	tr_rif_cache_entry *prev;
        long timestamp;
} tr_rif_dentry;


enum
{
	TR_USED=0,
	TR_UPDATED,
	TR_NUMDLISTS
};

struct tr_rif_cache_lists
{
	tr_rif_dlist       dlist[TR_NUMDLISTS];
	int                num_entries;
	int                hint_idx;
};

struct tr_rif_cache_entry  
{
	tr_rif_dentry      dentry[TR_NUMDLISTS];
	unsigned char      addr[TR_ALEN];
        TR_RIF_INFO
} __attribute__((packed));

#define RD_TO_RINGNUM(rd) (ntohs(rd)>>4)
#define RD_TO_BRIDGE_TYPE(rd)  (ntohs(rd)&0xf)

#ifdef __KERNEL__ 
void tr_free_rif_cache(struct net_device *dev);
#define TR_DEBUG_LOCKS    0
#if TR_DEBUG_LOCKS
extern int tr_lock2(struct device *dev,char *file,int line);
#define tr_lock(dev)      (tr_lock2(dev,__FILE__,__LINE__))
#else
#define tr_lock(dev)      (test_and_set_bit(1,&dev->tr_lock)==0)
#endif
#define tr_unlock(dev)    (atomic_set(&dev->tr_lock,0))
#endif /* __KERNEL__ */
#endif	/* _LINUX_IF_TR_H */
